;//******************************************************************************
;//* PROJECT:		Colour Palette Example
;//* AUTHOR:		Adrian Brown
;//* DATE:			25th May 2000
;//******************************************************************************
;//* HISTORY:
;//* 25/05/00		[APB]		Created initial example
;//******************************************************************************

INCLUDE			"Hardware.inc"

;// Wait for VRAM Macros
WaitForVRAM:	MACRO
				ld		a, [rSTAT] 
				and		a, STATF_OAM
				jr		nz, @-4
				ENDM

WaitForVRAMStore: MACRO
				push	af
				ld		a, [rSTAT] 
				and		a, STATF_OAM
				jr		nz, @-4
				pop		af
				ENDM

;// 
;// RGB Macro, allows using 0-255 range
;//

RGB:			MACRO
				DW		((\3 >> 3) << 10) + ((\2 >> 3) << 5) + (\1 >> 3)
				ENDM
				
;//******************************************************************************
;//*						INTERRUPTS HANDLERS									*
;//******************************************************************************

				SECTION "Interrupts", HOME[$0040]

				;// Since the space for the handlers is only 8 bytes - best jump to another location
VBInt:			jp		VBHandler

;//******************************************************************************
;//*						CARTRIDGE HEADER									*
;//******************************************************************************

				SECTION "Main", HOME[$0100]

				;//* [$0100] Program Start 0 Jump to main entry point
				nop
				jp		Initialise

				;//* [$0104] Cartridge Header - Nintendo logo etc
				NINTENDO_LOGO

				;// Cartridge name
				DB		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

				;// License code
				DB		0, 0

				;// Game type (0 = gameboy, 3 = super gameboy)
				DB		0

				;// Cartridge type
				DB		CART_ROM

				;// Rom Size
				DB		CART_ROM_256K

				;// Ram Size
				DB		CART_RAM_NONE

				;// Manufacturer code
				DW		0

				;// Version Number
				DB		0

				;// Complement check
				DB		$0a

				;// Checksum
				DW		0

;//******************************************************************************
;//*								MAIN ENTRY POINT							*
;//******************************************************************************

Initialise:		;// Set the stack pointer
				ld		sp, Stack

				;// Now disable interrupts and screen
				di

				;// Remember to wait for the VBlank
.WaitVB:		ld		a, [rLY]
				cp		a, 144
				jr		nz, .WaitVB

				;// Now disable everything
				ld		a, LCDCF_OFF
				ld		[rLCDC], a

				;// Clear the variables
				xor		a							
				ld		hl, MemoryStart				
				ld		bc, MemoryEnd - MemoryStart	
				call	Memory_Clear

				;// Clear video ram - we really should use a routine which checks for the video ram to be accessable
				;// but as the screen is off it will be - first clear the 1st bank

				ld		a, 1
				ld		[rVBK], a

				ld		hl, $8000
				ld		bc, $2000
				xor		a, a
				call	Memory_Clear

				;// Now clear bank 0
				xor		a, a
				ld		[rVBK], a

				ld		hl, $8000
				ld		bc, $2000
				call	Memory_Clear

				;// Upload the tile data
				ld		hl, TileData
				ld		de, $8800
				ld		c, 114
				call	MoveTilesToVideoRam

				;// Now upload the block map and attribute map - We should again wait for video ram to be available
				;// But the screen is off - so this is ok.
				ld		de, BlockMap
				ld		hl, $9800
				ld		b, 18
.YLoop:			ld		c, 20

.XLoop:			WaitForVRAM
				ld		a, [de]
				ld		[hl+], a
				inc		de
				dec		c
				jr		nz, .XLoop

				;// Remember to skip the extra 12 bytes as we are only doing one visible screen
				push	bc
				ld		bc, (32-20)
				add		hl, bc
				pop		bc

				dec		b
				jr		nz, .YLoop
				
				;// Swap ram banks
				ld		a, 1
				ld		[rVBK], a

				;// Attribute Map
				ld		de, AttribMap
				ld		hl, $9800
				ld		b, 18
.YALoop:		ld		c, 20

.XALoop:		WaitForVRAM
				ld		a, [de]
				ld		[hl+], a
				inc		de
				dec		c
				jr		nz, .XALoop

				;// Remember to skip the extra 12 bytes as we are only doing one visible screen
				push	bc
				ld		bc, (32-20)
				add		hl, bc
				pop		bc

				dec		b
				jr		nz, .YALoop

				;// Swap the ram banks back
				xor		a, a
				ld		[rVBK], a

				;// Set the background palettes
				ld		hl, Palette
				ld		b, 8
				xor		a, a

.PalLoop:		push	af
				push	bc
				call	SetPalette
				pop		bc
				pop		af

				inc		a
				dec		b
				jr		nz, .PalLoop

				;// Set Black Palette
				call	SetBlackPalette

				;// Initialise the fade
				ld		de, Palette
				ld		a, 8*4
				call	InitFade

				;// Enable the vertical blank interrupt
				ld		a, IEF_VBLANK
				ld		[rIE], a

				;// Enable the screen display
				ld		a, LCDCF_ON | LCDCF_WINOFF | LCDCF_BG8800 | LCDCF_BGON | LCDCF_BG9800 | LCDCF_OBJ8 | LCDCF_OBJOFF
				ld		[rLCDC], a

				;// Now call the entry point of the game after enabling the interrupts
				ei

				xor		a, a
				ld		[Visible], a

				;//
				;// Main Loop
				;//
Mainloop:		halt
				nop

				;// Was it a vertical blank
				ld		a, [VBFlag]
				or		a
				jr		z, Mainloop

				;// Clear the flag for next time				
				xor		a
				ld		[VBFlag], a

				;// Read the joypad
				call	Joypad_Read

				;// Did we press a button
				ld		a, [JP_New]
				and		a, a
				jr		z, .NoFade
				
				;// Fade the palette
				ld		a, $FF
				call	Fade

				;// Now setup the next fade
				ld		a, [Visible]
				and		a, a
				jr		z, .ToBlack

				xor		a, a
				ld		[Visible], a
				ld		de, Palette
				ld		a, 8*4
				call	InitFade
				jr		Mainloop

.ToBlack:
				ld		a, 1
				ld		[Visible], a
				ld		de, BlackPalette
				ld		a, 8*4
				call	InitFade
				jr		Mainloop

.NoFade:		;// Loop again
				jr		Mainloop

;//******************************************************************************
;//*								JOYPAD READER								*
;//******************************************************************************
Joypad_Read:	push	bc							;// Store the registers
				push	hl

				ld		hl, rP1						;// Set HL to be the Joypad address
				ld		a, P1F_4					;// Select buttons
				ld		[hl], a
				ld		a, [hl]						;// Delay the read
				ld		a, [hl]						;// Get the values
				cpl
				and		a, $0F
				swap	a
				ld		b, a						;// Get the buttons in the top bits and store in B
				
				ld		a, P1F_5					;// Select directions
				ld		[hl], a
				ld		a, [hl]
				ld		a, [hl]
				ld		a, [hl]
				ld		a, [hl]
				ld		a, [hl]
				ld		a, [hl]
				cpl
				and		$0F							;// Get the directions
				or		b							;// Mix the two together

				ld		b, a						;// Now store the value
				ld		a, [JP_Old]					;// Get the old value
				xor		b
				and		b							;// Get the current bits
				ld		[JP_New], a					;// Store the new bits
				ld		a, b
				ld		[JP_Old], a					;// Restore the old values

				ld		a, P1F_4 | P1F_5
				ld		[hl], a						;// Reset the joypad controller

				pop		hl							;// Restore the registers
				pop		bc
				ret
				
;//******************************************************************************
;//*								VERTICAL BLANK HANDLING						*
;//******************************************************************************

				;// The first thing that needs to be done is registers stored
VBHandler:		push	af
				push	hl

				;// Set a flag so we know when this has happened
				ld		a, 1
				ld		[VBFlag], a

				;// Finally - restore the registers and reti to return
				pop		hl
				pop		af
				reti

;//******************************************************************************
;//*								MISC FUNCTIONS								*
;//******************************************************************************
;// Memory Clear - Clears a chunk of memory
;//	ENTRY:
;//  A  = Value to clear to
;//  HL = Address of memory
;//  BC = Length of memory
Memory_Clear:
				;// Increase the values so we are one based
				inc		b
				inc		c
				jr		.EndCheck

				;// Clear the bytes
.ClearLoop:		ld		[hl+], a
.EndCheck:		dec		c
				jr		nz, .ClearLoop

				;// Check the high byte
				dec		b
				jr		nz, .ClearLoop

				;// Exit
				ret
				
;// Memory Copy - Copys a chunk of memory - This should not be used to move into video ram unless the screen 
;//				  is disabled as it does not wait for video ram
;//	ENTRY:
;//  HL = Source Address
;//	 DE = Destination Address
;//  BC = Length of memory
Memory_Copy:
				;// Increase the values so we are one based
				inc		b
				inc		c
				jr		.EndCheck

				;// Clear the bytes
.ClearLoop:		ld		a, [hl+]
				ld		[de], a
				inc		de
.EndCheck:		dec		c
				jr		nz, .ClearLoop

				;// Check the high byte
				dec		b
				jr		nz, .ClearLoop

				;// Exit
				ret
				
;// MoveToVideoRam - Moves a chunk of tiles to video ram
;//	ENTRY:
;//  C = Number of tiles
;//  HL = Address of Tile Data
;//  DE = Address to move to
MoveTilesToVideoRam:
.TileLoop:
				;// 16 Bytes per tile
				ld		b, 16

.DataLoop:
				;// Wait for video ram
				WaitForVRAM

				ld		a, [hl+]
				ld		[de], a
				inc		de
				dec		b
				jr		nz, .DataLoop										;// Loop for all 16 bytes

				dec		c
				jr		nz, .TileLoop										;// Loop for all the tiles

				ret

;// SetPalette - Sets a background palette
;//	ENTRY:
;//  A = Palette to set
;//  HL = Palette Address
SetPalette:
				add		a, a												; Get the correct palette to set (*8)
				add		a, a
				add		a, a
				add		a, $80												; Automatically increase address
				ld		[$FF68], a

				ld		b, 8												; 8 bytes per palette
				ld		c, $69												; Write Data register

.Loop:
				WaitForVRAM
				ld		a, [hl+]
				ld		[c], a
				dec		b
				jr		nz, .Loop

				ret

;// SetBlackPalette - Sets a background palette to black
SetBlackPalette:
				ld		a, $80
				ld		[$FF68], a

				ld		bc, $4069
.Loop:			WaitForVRAM
				xor		a, a
				ld		[c], a
				dec		b
				jr		nz, .Loop

				ret

;// ConvertTo24Bit - Converts a palette word into a 24 bit colour
;//	ENTRY:
;//  BC = Palette Word
;// OUT:
;//  E  = Red Componenet
;//  C  = Green Component
;//  B  = Blue Componenet
ConvertTo24Bit:	
				push	af

				;// Get the Red value in E
				ld		a, c
				and		a, $1f
				ld		e, a

				;// Shift out the Blue
				srl     b
				rr      c
				srl     b
				rr      c

				;// Get the Green
		        srl     c
		        srl     c
		        srl     c

				pop		af
				ret

;// InitFade - Initialises the data ready for a fade
;//	ENTRY:
;//  DE = Palette Address of Fade to Palette
;//  A	= Number of palettes
InitFade:		
				;// First convert the palettes to the 24bit versions
				ld		[PAL_NumberOfPals], a
				ld		b, a
				push	bc
				push	de

				;//
				;// Get the current palettes
				;//
				ld		hl, PAL_StartingPal
				ld		e, $80

.CurrentLoop:	push	bc

				;// Get the current low byte
				WaitForVRAM
				ld		a, e
				inc		e
				ld		[$FF68], a
				ld		a, [$FF69]
				ld		c, a

				;// Read the high byte
				WaitForVRAM
				ld		a, e
				inc		e
				ld		[$FF68], a
				ld		a, [$FF69]
				ld		b, a

				;// Store the read value
				push	de

				xor		a, a

				;// Convert to 24 bit
				Call	ConvertTo24Bit

				;// Store Red in 8:8 format
				ld		[hl+], a
				ld		[hl], e
				inc		hl

				;// Store Green in 8:8 format
				ld		[hl+], a
				ld		[hl], c
				inc		hl

				;// Store Blue in 8:8 format
				ld		[hl+], a
				ld		[hl], b
				inc		hl

				pop		de
				pop		bc
				dec		b
				jr		nz, .CurrentLoop

				;// Restore the pointers and counts
				pop		de
				pop		bc

				;// Get the store address
				ld		hl, PAL_PaletteDiff

				;// Start the reading
				ld		a, $80
				ld		[PAL_TempStore], a

				;//
				;// Calculate the differences
				;//
.CalcLoop:		push	bc

				;// Get the wanted
				ld		a, [de]
				inc		de
				ld		c, a
				ld		a, [de]
				inc		de
				ld		b, a

				;// Store DE
				push	de

				;// Convert the colours
				call	ConvertTo24Bit

				;// Store the Green and Blue
				push	bc

				;// Store Address of differences
				push	hl

				;// Get the current value
				WaitForVRAM
				ld		a, [PAL_TempStore]
				ld		l, a
				inc		l
				ld		[$FF68], a
				ld		a, [$FF69]
				ld		c, a

				WaitForVRAM
				ld		a, l
				inc		l
				ld		[$FF68], a
				ld		a, l
				ld		[PAL_TempStore], a
				ld		a, [$FF69]
				ld		b, a

				;// Store Red in A
				ld		a, e
				call	ConvertTo24Bit

				;//
				;// CALCULATE THE RED DIFFERENCE - ENDS UP IN E:A 8:8 fixed value
				;//
				sub		a, e
				ld		e, 0
				jr		nc, .RedNotNeg
				dec		e
.RedNotNeg:

				;// Divide by 32 (done by shifting left by 3 as its quicker than right by 5, remember the bytes
				;// will be the wrong way around)
				sla		a
				rl		e
				sla		a
				rl		e
				sla		a
				rl		e

				;// Get the current write address
				pop		hl

				;// At this point:
				;//  A = Current Low byte of difference
				;//  B = Blue Current Colour Value
				;//  C = Green Current Colour Value
				;//  D = 
				;//  E = High byte of difference
				;//  HL = Address of store

				;// Now store difference
				ld		[hl+], a
				ld		[hl], e
				inc		hl

				;// Get the Green and Blue back
				pop		de

				;//
				;// CALCULATE THE GREEN DIFFERENCE - ENDS UP IN C:A 8:8 fixed value
				;//
				ld		a, e
				sub		c
				ld		c, 0
				jr		nc, .GreenNotNeg
				dec		c
.GreenNotNeg:

				;// Do the divide again
				sla		a
				rl		c
				sla		a
				rl		c
				sla		a
				rl		c

				;// Now store the green difference
				ld		[hl+], a
				ld		[hl], c
				inc		hl

				;//
				;// CALCULATE THE BLUE DIFFERENCE - ENDS UP IN B:A 8:8 fixed value
				;//
				ld		a, d
				sub		b
				ld		b, 0
				jr		nc, .BlueNotNeg
				dec		b
.BlueNotNeg:

				;// Do the divide again
				sla		a
				rl		b
				sla		a
				rl		b
				sla		a
				rl		b

				;// Now store the blue difference
				ld		[hl+], a
				ld		[hl], b
				inc		hl

				;// Loop for all the colours
				pop		de
				pop		bc
				dec		b
				jr		nz, .CalcLoop

				ret

;// Fade - Fades to the palette set in InitFade over 32 frames
;//	ENTRY:
;//  A	= Mask for buttons to skip with
Fade:			push	af

				;// Loop for 32
				ld		b, 32
.FadeLoop:		push	bc

				;// Get the pointers
				ld		de, PAL_StartingPal
				ld		hl, PAL_PaletteDiff
				ld		bc, PAL_ActualPals

				;// Loop through all the paletes
				ld		a, [PAL_NumberOfPals]

.PaletteLoop:	push	af
				push	bc

				;// Increase the byte
				ld		a, [de]
				add		a, [hl]
				ld		[de], a
				inc		hl
				inc		de
				ld		a, [de]
				adc		a, [hl]
				ld		[de], a
				inc		hl
				inc		de

				;// Store Red in B
				ld		b, a

				;// Increase the byte
				ld		a, [de]
				add		a, [hl]
				ld		[de], a
				inc		hl
				inc		de
				ld		a, [de]
				adc		a, [hl]
				ld		[de], a
				inc		hl
				inc		de

				;// Store Green in C
				ld		c, a

				;// Increase the byte
				ld		a, [de]
				add		a, [hl]
				ld		[de], a
				inc		hl
				inc		de
				ld		a, [de]
				adc		a, [hl]
				ld		[de], a
				inc		hl
				inc		de

				;// At this point:
				;//  A = Blue Colour
				;//  B = Red Colour
				;//  C = Green Colour
				;//  DE = Starting Palette
				;//  HL = Difference Address

				;// Convert to 15 bit - Shift up so we can shift the green bits in
				rlc		b
				rlc		b
				rlc		b

				;// Shift the screen bits in
				srl		c
				rr		b
				srl		c
				rr		b
				srl		c
				rr		b

				;// Finally get the blue bits in
				add		a, a
				add		a, a
				add		a, c

				;// A = High byte, B = Low Byte
				ld		[PAL_TempStore], a
				ld		a, b

				pop		bc
				ld		[bc], a
				inc		bc

				ld		a, [PAL_TempStore]
				ld		[bc], a
				inc		bc

				pop		af
				dec		a
				jr		nz, .PaletteLoop

.FadeLoopVB		halt
				nop

				;// Was it a vertical blank
				ld		a, [VBFlag]
				or		a
				jr		z, .FadeLoopVB

				;// Clear the flag for next time				
				xor		a
				ld		[VBFlag], a

				;// We want to write to all the palettes
				ld		a, $80
				ld		[$FF68], a

				ld		hl, PAL_ActualPals

				;// Remember 2 bytes per palette
				ld		a, [PAL_NumberOfPals]
				add		a, a
				ld		b, a
				ld		c, $69

.SetPalette:	ld		a, [hl+]
				ld		[c], a
				dec		b
				jr		nz, .SetPalette

				pop		bc
				dec		b
				jr		nz, .FadeLoop

				pop		af
				ret

;//*********************************************************************************************************************
;// TILE DATA
;//*********************************************************************************************************************
TileData:
				INCLUDE	"TileData.ASM"

BlockMap:
				INCLUDE	"BlockMap.ASM"

AttribMap:
				INCLUDE	"AttribMap.ASM"

Palette:
				;//RGB		255, 239, 135
				RGB		255, 255, 255
				RGB		223, 159, 191
				RGB		  0,  87, 191
				RGB		159, 191, 247

				RGB		  0, 151, 159
				RGB		159, 231, 255
				RGB		  0,  87, 191
				RGB		255, 255, 255

				RGB		  0,  55,  87
				RGB		127, 135, 199
				RGB		  0,  87, 191
				RGB		159, 231, 255

				RGB		  0, 119, 175
				RGB		159, 231, 255
				RGB		223, 159, 191
				RGB		255, 255, 255

				RGB		 63, 151, 151
				RGB		  0,  87, 191
				RGB		  0, 151, 159
				RGB		159, 231, 255

				RGB		 63, 151, 151
				RGB		223, 159, 191
				RGB		  0,  87, 191
				RGB		111, 255, 255

				RGB		255, 239, 135
				RGB		127, 135, 199
				RGB		  0,  71, 135
				RGB		159, 231, 255

				RGB		 25,  25,  25
				RGB		  0,  87, 191
				RGB		223, 143,  55
				RGB		255, 255, 255

BlackPalette:	
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0
				RGB		  0,   0,   0

;//******************************************************************************
;//*								VARIABLES									*
;//******************************************************************************
				SECTION "Memory", BSS[$C000]

MemoryStart:
;//* INTERRUPT VARIABLES
VBFlag:			DS		1
JP_Old:			DS		1
JP_New:			DS		1
Visible:		DS		1

;//* PALETTE FADE VARIABLES
PAL_NumberOfPals:	DS	1
PAL_TempStore:		DS	1
PAL_ActualPals:		DS	2*4*8
PAL_StartingPal:	DS	6*4*8
PAL_PaletteDiff:	DS	6*4*8
MemoryEnd:

;//******************************************************************************
;//*							STACK SPACE										*
;//******************************************************************************

				DS		32 * 2
Stack: